@@ -37,13 +37,12 @@
 
 /* careful, this file includes data *declarations* */
 #include "sparc-opcode.h"
+#include "stuff/target_arch.h"
 #include <mach-o/sparc/reloc.h>
 
 /* From GNU ansidecl.h */
 #define PARAMS(paramlist)		paramlist
 
-typedef long	offsetT;
-
 /*
  * These are the default cputype and cpusubtype for the Sparc architecture.
  */
@@ -84,7 +83,6 @@ static struct hash_control *op_hash = NULL;
 
 #ifdef	NeXT_MOD
 static void s_proc PARAMS ((int));
-static void s_ignore PARAMS ((int));
 extern void s_seg PARAMS ((int));
 #else	/* NeXT_MOD */
 static void s_data1 PARAMS ((void));
@@ -165,7 +163,7 @@ struct sparc_it
   {
     char *error;
     unsigned long opcode;
-    struct nlist *nlistp;
+    nlist_t *nlistp;
     expressionS exp;
     int pcrel;
     char pcrel_reloc;	/* do relocation? */
@@ -207,14 +205,6 @@ s_proc (ignore)
   totally_ignore_line();
 }
 
-/* we simply ignore the rest of this statement */
-static void
-s_ignore (ignore)
-     int ignore;
-{
-  totally_ignore_line();
-}
-
 /* This function is called once, at assembler startup time.  It should
    set up all the tables, etc. that the MD part of the assembler will need. */
 void
@@ -222,7 +212,7 @@ md_begin ()
 {
   register const char *retval = NULL;
   int lose = 0;
-  register unsigned int i = 0;
+  register int i = 0;
 
   op_hash = hash_new ();
 
@@ -992,7 +982,7 @@ sparc_ip (str)
       if (match == 0)
 	{
 	  /* Args don't match. */
-	  if (((unsigned) (&insn[1] - sparc_opcodes)) < NUMOPCODES
+	  if (((int) (&insn[1] - sparc_opcodes)) < NUMOPCODES
 	      && !strcmp (insn->name, insn[1].name))
 	    {
 	      ++insn;
@@ -1141,7 +1131,7 @@ md_atof (type, litP, sizeP)
 void
 md_number_to_chars (buf, val, n)
      char *buf;
-     long val;
+     signed_expr_t val;
      int n;
 {
   // sigh, all architectures do this..,
@@ -1166,7 +1156,7 @@ md_number_to_chars (buf, val, n)
 
 
 void
-md_number_to_imm(unsigned char *buf, long val, int size, fixS *fixP, int nsect)
+md_number_to_imm(unsigned char *buf, signed_expr_t val, int size, fixS *fixP, int nsect)
 {
 
   /* handle the most common case quickly */
