@@ -1,11 +1,14 @@
 #include <stdlib.h>
 #include <string.h>
+#include <ctype.h>
 #include <sys/file.h>
 #include <libc.h>
 #include <mach/mach.h>
+#include "stuff/target_arch.h"
 #include "stuff/openstep_mach.h"
 #include <mach-o/loader.h>
 #include <mach-o/reloc.h>
+#include <mach-o/stab.h>
 #ifdef I860
 #include <mach-o/i860/reloc.h>
 #endif
@@ -22,6 +25,9 @@
 #ifdef SPARC
 #include <mach-o/sparc/reloc.h>
 #endif
+#if defined(I386) && defined(ARCH64)
+#include <mach-o/x86_64/reloc.h>
+#endif
 #include "stuff/round.h"
 #include "stuff/bytesex.h"
 #include "stuff/errors.h"
@@ -34,29 +40,37 @@
 #include "sections.h"
 #include "messages.h"
 #include "xmalloc.h"
+#include "input-scrub.h"
 #ifdef I860
-#define RELOC_SECTDIFF	I860_RELOC_SECTDIFF
-#define RELOC_PAIR	I860_RELOC_PAIR
+#define RELOC_SECTDIFF		I860_RELOC_SECTDIFF
+#define RELOC_LOCAL_SECTDIFF	I860_RELOC_SECTDIFF
+#define RELOC_PAIR		I860_RELOC_PAIR
 #endif
 #ifdef M88K
-#define RELOC_SECTDIFF	M88K_RELOC_SECTDIFF
-#define RELOC_PAIR	M88K_RELOC_PAIR
+#define RELOC_SECTDIFF		M88K_RELOC_SECTDIFF
+#define RELOC_LOCAL_SECTDIFF	M88K_RELOC_SECTDIFF
+#define RELOC_PAIR		M88K_RELOC_PAIR
 #endif
 #ifdef PPC
-#define RELOC_SECTDIFF	PPC_RELOC_SECTDIFF
-#define RELOC_PAIR	PPC_RELOC_PAIR
+#define RELOC_SECTDIFF		PPC_RELOC_SECTDIFF
+#define RELOC_LOCAL_SECTDIFF	PPC_RELOC_LOCAL_SECTDIFF
+#define RELOC_PAIR		PPC_RELOC_PAIR
+#define PPC_RELOC_BR14_predicted (0x10 | PPC_RELOC_BR14)
 #endif
 #ifdef HPPA
-#define RELOC_SECTDIFF	HPPA_RELOC_SECTDIFF
-#define RELOC_PAIR	HPPA_RELOC_PAIR
+#define RELOC_SECTDIFF		HPPA_RELOC_SECTDIFF
+#define RELOC_LOCAL_SECTDIFF	HPPA_RELOC_SECTDIFF
+#define RELOC_PAIR		HPPA_RELOC_PAIR
 #endif
 #ifdef SPARC
-#define RELOC_SECTDIFF	SPARC_RELOC_SECTDIFF
-#define RELOC_PAIR	SPARC_RELOC_PAIR
+#define RELOC_SECTDIFF		SPARC_RELOC_SECTDIFF
+#define RELOC_LOCAL_SECTDIFF	SPARC_RELOC_SECTDIFF
+#define RELOC_PAIR		SPARC_RELOC_PAIR
 #endif
 #if defined(M68K) || defined(I386)
-#define RELOC_SECTDIFF	GENERIC_RELOC_SECTDIFF
-#define RELOC_PAIR	GENERIC_RELOC_PAIR
+#define RELOC_SECTDIFF		GENERIC_RELOC_SECTDIFF
+#define RELOC_LOCAL_SECTDIFF	GENERIC_RELOC_LOCAL_SECTDIFF
+#define RELOC_PAIR		GENERIC_RELOC_PAIR
 #endif
 
 /*
@@ -110,8 +124,8 @@ write_object(
 char *out_file_name)
 {
     /* The structures for Mach-O relocatables */
-    struct mach_header		header;
-    struct segment_command	reloc_segment;
+    mach_header_t		header;
+    segment_command_t		reloc_segment;
     struct symtab_command	symbol_table;
     struct dysymtab_command	dynamic_symbol_table;
     unsigned long		section_type, *indirect_symbols;
@@ -139,6 +153,8 @@ char *out_file_name)
     long num_bytes;
     char *symbol_name;
     int fd;
+    unsigned long local;
+    struct stat stat_buf;
 
 #ifdef I860
 	I860_tweeks();
@@ -179,7 +195,7 @@ char *out_file_name)
 	layout_symbols((long *)&nsyms, (long *)&strsize);
 
 	/* fill in the Mach-O header */
-	header.magic = MH_MAGIC;
+	header.magic = MH_MAGIC_VALUE;
 	header.cputype = md_cputype;
 	if(archflag_cpusubtype != -1)
 	    header.cpusubtype = archflag_cpusubtype;
@@ -191,8 +207,8 @@ char *out_file_name)
 	header.sizeofcmds = 0;
 	if(nsects != 0){
 	    header.ncmds += 1;
-	    header.sizeofcmds += sizeof(struct segment_command) +
-				 nsects * sizeof(struct section);
+	    header.sizeofcmds += sizeof(segment_command_t) +
+				 nsects * sizeof(section_t);
 	}
 	if(nsyms != 0){
 	    header.ncmds += 1;
@@ -205,17 +221,22 @@ char *out_file_name)
 	else
 	    strsize = 0;
 	header.flags = 0;
+	if(subsections_via_symbols == TRUE)
+	    header.flags |= MH_SUBSECTIONS_VIA_SYMBOLS;
+#ifdef ARCH64
+	header.reserved = 0;
+#endif
 
 	/* fill in the segment command */
-	memset(&reloc_segment, '\0', sizeof(struct segment_command));
-	reloc_segment.cmd = LC_SEGMENT;
-	reloc_segment.cmdsize = sizeof(struct segment_command) +
-				nsects * sizeof(struct section);
+	memset(&reloc_segment, '\0', sizeof(segment_command_t));
+	reloc_segment.cmd = LC_SEGMENT_VALUE;
+	reloc_segment.cmdsize = sizeof(segment_command_t) +
+				nsects * sizeof(section_t);
 	/* leave reloc_segment.segname full of zeros */
 	reloc_segment.vmaddr = 0;
 	reloc_segment.vmsize = 0;
 	reloc_segment.filesize = 0;
-	offset = header.sizeofcmds + sizeof(struct mach_header);
+	offset = header.sizeofcmds + sizeof(mach_header_t);
 	reloc_segment.fileoff = offset;
 	reloc_segment.maxprot = VM_PROT_READ | VM_PROT_WRITE | VM_PROT_EXECUTE;
 	reloc_segment.initprot= VM_PROT_READ | VM_PROT_WRITE | VM_PROT_EXECUTE;
@@ -322,7 +343,7 @@ char *out_file_name)
 	else
 	    symbol_table.symoff = offset;
 	symbol_table.nsyms = nsyms;
-	offset += symbol_table.nsyms * sizeof(struct nlist);
+	offset += symbol_table.nsyms * sizeof(nlist_t);
 
 	/* fill in the string table fields of the symtab_command */
 	if(strsize == 0)
@@ -352,31 +373,31 @@ char *out_file_name)
 	offset = 0;
 
 	/* put the mach_header in the buffer */
-	memcpy(output_addr + offset, &header, sizeof(struct mach_header));
+	memcpy(output_addr + offset, &header, sizeof(mach_header_t));
 	if(host_byte_sex != md_target_byte_sex)
-	    swap_mach_header((struct mach_header *)(output_addr + offset),
-			     md_target_byte_sex);
-	offset += sizeof(struct mach_header);
+	    swap_mach_header_t((mach_header_t *)(output_addr + offset),
+			       md_target_byte_sex);
+	offset += sizeof(mach_header_t);
 
 	/* put the segment_command in the buffer */
 	if(nsects != 0){
 	    memcpy(output_addr + offset, &reloc_segment,
-		   sizeof(struct segment_command));
+		   sizeof(segment_command_t));
 	    if(host_byte_sex != md_target_byte_sex)
-		swap_segment_command((struct segment_command *)
-				     (output_addr + offset),
-				     md_target_byte_sex);
-	    offset += sizeof(struct segment_command);
+		swap_segment_command_t((segment_command_t *)
+				       (output_addr + offset),
+				       md_target_byte_sex);
+	    offset += sizeof(segment_command_t);
 	}
 
 	/* put the segment_command's section structures in the buffer */
 	for(frchainP = frchain_root; frchainP; frchainP = frchainP->frch_next){
 	    memcpy(output_addr + offset, &(frchainP->frch_section),
-		   sizeof(struct section));
+		   sizeof(section_t));
 	    if(host_byte_sex != md_target_byte_sex)
-		swap_section((struct section *)(output_addr + offset), 1,
-				     md_target_byte_sex);
-	    offset += sizeof(struct section);
+		swap_section_t((section_t *)(output_addr + offset), 1,
+			       md_target_byte_sex);
+	    offset += sizeof(section_t);
 	}
 
 	/* put the symbol_command in the buffer */
@@ -415,19 +436,8 @@ char *out_file_name)
 		/* put the variable repeated part of the frag in the buffer */
 		fill_literal = fragP->fr_literal + fragP->fr_fix;
 		fill_size = fragP->fr_var;
-		know(fragP->fr_offset >= 0);
-		if(fill_size != 0)
-		    num_bytes = fragP->fr_offset % fill_size;
-		else
-		    num_bytes = 0;
-		if (num_bytes){
-		    long zero = 0;
-		    memcpy(output_addr + offset, &zero, num_bytes);
-		    offset += num_bytes;
-		}
-		for(count = fragP->fr_offset-num_bytes;
-		    count > 0;
-		    count -= fill_size){
+		num_bytes = fragP->fr_offset * fragP->fr_var;
+		for(count = 0; count < num_bytes; count += fill_size){
 		    memcpy(output_addr + offset, fill_literal, fill_size);
 		    offset += fill_size;
 		}
@@ -439,41 +449,54 @@ char *out_file_name)
 	offset = symbol_table.symoff;
 	for(symbolP = symbol_rootP; symbolP; symbolP = symbolP->sy_next){
 	    if((symbolP->sy_type & N_EXT) == 0){
-		symbol_name = symbolP->sy_nlist.n_un.n_name;
+		symbol_name = symbolP->sy_name;
 		symbolP->sy_nlist.n_un.n_strx = symbolP->sy_name_offset;
 		if(symbolP->expression != 0) {
 		    expressionS *exp;
 
 		    exp = (expressionS *)symbolP->expression;
+		    if((exp->X_add_symbol->sy_type & N_TYPE) == N_UNDF)
+	    		as_fatal("undefined symbol `%s' in operation setting "
+				 "`%s'", exp->X_add_symbol->sy_name,
+				 symbol_name);
+		    if((exp->X_subtract_symbol->sy_type & N_TYPE) == N_UNDF)
+	    		as_fatal("undefined symbol `%s' in operation setting "
+				 "`%s'", exp->X_subtract_symbol->sy_name,
+				 symbol_name);
+		    if(exp->X_add_symbol->sy_other !=
+		       exp->X_subtract_symbol->sy_other)
+	    		as_fatal("invalid sections for operation on `%s' and "
+				 "`%s' setting `%s'",exp->X_add_symbol->sy_name,
+				 exp->X_subtract_symbol->sy_name, symbol_name);
 		    symbolP->sy_nlist.n_value +=
 			exp->X_add_symbol->sy_value -
 			exp->X_subtract_symbol->sy_value;
 		}
 		memcpy(output_addr + offset, (char *)(&symbolP->sy_nlist),
-		       sizeof(struct nlist));
-		symbolP->sy_nlist.n_un.n_name = symbol_name;
-		offset += sizeof(struct nlist);
+		       sizeof(nlist_t));
+		symbolP->sy_name = symbol_name;
+		offset += sizeof(nlist_t);
 	    }
 	}
 	for(i = 0; i < nextdefsym; i++){
-	    symbol_name = extdefsyms[i]->sy_nlist.n_un.n_name;
+	    symbol_name = extdefsyms[i]->sy_name;
 	    extdefsyms[i]->sy_nlist.n_un.n_strx = extdefsyms[i]->sy_name_offset;
 	    memcpy(output_addr + offset, (char *)(&extdefsyms[i]->sy_nlist),
-	           sizeof(struct nlist));
-	    extdefsyms[i]->sy_nlist.n_un.n_name = symbol_name;
-	    offset += sizeof(struct nlist);
+	           sizeof(nlist_t));
+	    extdefsyms[i]->sy_name = symbol_name;
+	    offset += sizeof(nlist_t);
 	}
 	for(j = 0; j < nundefsym; j++){
-	    symbol_name = undefsyms[j]->sy_nlist.n_un.n_name;
+	    symbol_name = undefsyms[j]->sy_name;
 	    undefsyms[j]->sy_nlist.n_un.n_strx = undefsyms[j]->sy_name_offset;
 	    memcpy(output_addr + offset, (char *)(&undefsyms[j]->sy_nlist),
-	           sizeof(struct nlist));
-	    undefsyms[j]->sy_nlist.n_un.n_name = symbol_name;
-	    offset += sizeof(struct nlist);
+	           sizeof(nlist_t));
+	    undefsyms[j]->sy_name = symbol_name;
+	    offset += sizeof(nlist_t);
 	}
 	if(host_byte_sex != md_target_byte_sex)
-	    swap_nlist((struct nlist *)(output_addr + symbol_table.symoff),
-		       symbol_table.nsyms, md_target_byte_sex);
+	    swap_nlist_t((nlist_t *)(output_addr + symbol_table.symoff),
+		         symbol_table.nsyms, md_target_byte_sex);
 
 	/*
 	 * Put the relocation entries for each section in the buffer.
@@ -508,10 +531,30 @@ char *out_file_name)
 		    for(isymbolP = frchainP->frch_isym_root;
 			isymbolP != NULL;
 			isymbolP = isymbolP->isy_next){
-
-			memcpy(output_addr + offset,
-			       (char *)(&isymbolP->isy_symbol->sy_number),
-			       sizeof(unsigned long));
+			/*
+			 * If this is a non-lazy pointer symbol section and
+			 * if the symbol is a local symbol then put out
+			 * INDIRECT_SYMBOL_LOCAL as the indirect symbol table
+			 * entry.  This is used with code gen for fix-n-continue
+			 * where the compiler generates indirection for static
+			 * data references.  See the comments at the end of
+			 * fixup_section() that explains the assembly code used.
+			 */
+			if(section_type == S_NON_LAZY_SYMBOL_POINTERS &&
+			   (isymbolP->isy_symbol->sy_nlist.n_type & N_EXT) !=
+			    N_EXT){
+    			    local = INDIRECT_SYMBOL_LOCAL;
+			    if((isymbolP->isy_symbol->sy_nlist.n_type &
+				N_TYPE) == N_ABS)
+				local |= INDIRECT_SYMBOL_ABS;
+			    memcpy(output_addr + offset, (char *)(&local),
+	   			   sizeof(unsigned long));
+			}
+			else{
+			    memcpy(output_addr + offset,
+				   (char *)(&isymbolP->isy_symbol->sy_number),
+				   sizeof(unsigned long));
+			}
 			offset += sizeof(unsigned long);
 		    }
 		}
@@ -559,10 +602,17 @@ char *out_file_name)
          */
 	if(bad_error != 0)
 	    return;
-	(void)unlink(out_file_name);
+	/*
+	 * Avoid doing the unlink() on special files, just unlink regular files
+	 * that exist.
+	 */
+	if(stat(out_file_name, &stat_buf) != -1){
+	    if(stat_buf.st_mode & S_IFREG)
+		(void)unlink(out_file_name);
+	}
 	if((fd = open(out_file_name, O_WRONLY | O_CREAT | O_TRUNC, 0666)) == -1)
 	    as_fatal("can't create output file: %s", out_file_name);
-	if(write(fd, output_addr, output_size) != output_size)
+	if(write(fd, output_addr, output_size) != (int)output_size)
 	    as_fatal("can't write output file");
 	if(close(fd) == -1)
 	    as_fatal("can't close output file");
@@ -653,9 +703,9 @@ symbolP = symbol_find_or_make(isymbolP->isy_name);
 		if(section_type == S_SYMBOL_STUBS)
 		    stride = frchainP->frch_section.reserved2;
 		else
-		    stride = sizeof(unsigned long);
+		    stride = sizeof(signed_target_addr_t);
 		if(frchainP->frch_section.size / stride != count)
-		    as_warn("missing indirect symbols for section (%s,%s)",
+		    as_bad("missing indirect symbols for section (%s,%s)",
 			    frchainP->frch_section.segname,
 			    frchainP->frch_section.sectname);
 		/*
@@ -669,6 +719,96 @@ symbolP = symbol_find_or_make(isymbolP->isy_name);
 	return(total);
 }
 
+
+/*
+ * set_BINCL_checksums() walks through all STABS and calculate BINCL checksums. This will improve
+ * linking performance because the linker will not need to touch and sum STABS
+ * strings to do the BINCL/EINCL duplicate removal.
+ *
+ * A BINCL checksum is a sum of all stabs strings within a BINCL/EINCL pair.
+ * Since BINCL/EINCL can be nested, a stab string contributes to only the 
+ * innermost BINCL/EINCL enclosing it. 
+ *
+ * The checksum excludes the first number after an open paren.
+ *
+ * Some stabs (e.g. SLINE) when found within a BINCL/EINCL disqualify the EXCL
+ * optimization and therefore disable this checksumming.
+ */
+static
+void
+set_BINCL_checksums()
+{
+    struct HeaderRange { 
+	symbolS*		bincl; 
+	struct HeaderRange*	parentRange;
+	unsigned int		sum;
+	int			okToChecksum; 
+    };
+    symbolS *symbolP;
+    struct HeaderRange* curRange = NULL;
+	
+	for(symbolP = symbol_rootP; symbolP; symbolP = symbolP->sy_next){
+	    if((symbolP->sy_nlist.n_type & N_STAB) != 0){
+		switch(symbolP->sy_nlist.n_type){
+		case N_BINCL:
+		    {
+			struct HeaderRange* range =
+			    xmalloc(sizeof(struct HeaderRange));
+			range->bincl = symbolP;
+			range->parentRange = curRange;
+			range->sum = 0; 
+			range->okToChecksum = (symbolP->sy_nlist.n_value == 0);
+			curRange = range;
+		    }
+		    break;
+		case N_EINCL:
+		    if(curRange != NULL){
+			struct HeaderRange* tmp = curRange;
+			if (curRange->okToChecksum)
+			    curRange->bincl->sy_nlist.n_value = curRange->sum;
+			curRange = tmp->parentRange;
+			free(tmp);
+		    }
+		    break;				
+		case N_FUN:
+		case N_BNSYM:
+		case N_ENSYM:
+		case N_LBRAC:
+		case N_RBRAC:
+		case N_SLINE:
+		case N_STSYM:
+		case N_LCSYM:
+		    if(curRange != NULL){
+			curRange->okToChecksum = FALSE;
+		    }
+		    break;
+		case N_EXCL:
+			break;
+		default:
+		    if(curRange != NULL){
+			if(curRange->okToChecksum){
+			    unsigned int sum = 0;
+			    const char* s = symbolP->sy_name;
+			    char c;
+			    while((c = *s++) != '\0'){
+				sum += c;
+				/*
+				 * Don't checkusm first number (file index)
+				 * after open paren in string.
+				 */
+				if(c == '('){
+				    while(isdigit(*s))
+					++s;
+				}
+			    }
+			    curRange->sum += sum;
+			}
+		    }
+		}
+	    }
+	}
+}
+
 /*
  * layout_symbols() removes temporary symbols (symbols that are of the form L1
  * and 1:) if the -L flag is not seen so the symbol table has only the symbols
@@ -693,6 +833,7 @@ long *string_byte_count)
     symbolS *symbolP;
     symbolS **symbolPP;
     char *name;
+	int seenBINCL = FALSE;
 
 	*symbol_number = 0;
 	*string_byte_count = sizeof(char);
@@ -818,6 +959,9 @@ long *string_byte_count)
 		    /* the ordinary case (symbol has a name) */
 		    symbolP->sy_name_offset = *string_byte_count;
 		    *string_byte_count += strlen(symbolP->sy_name) + 1;
+		    /* check for existance of BINCL/EINCL */
+		    if(symbolP->sy_nlist.n_type == N_BINCL)
+			seenBINCL = TRUE;
 		}
 		else{
 		    /* the .stabd case (symbol has no name) */
@@ -843,6 +987,10 @@ long *string_byte_count)
 	    undefsyms[j]->sy_number = *symbol_number;
 	    *symbol_number = *symbol_number + 1;
 	}
+	
+	/* calculate BINCL checksums */
+	if(seenBINCL)
+	    set_BINCL_checksums();	
 }
 
 /*
@@ -944,6 +1092,10 @@ struct relocation_info *riP)
 	if(fixP->fx_addsy == NULL)
 	    return(0);
 
+#ifdef TC_VALIDATE_FIX
+	TC_VALIDATE_FIX(fixP, sect_addr, 0);
+#endif
+
 	memset(riP, '\0', sizeof(struct relocation_info));
 	symbolP = fixP->fx_addsy;
 
@@ -955,9 +1107,21 @@ struct relocation_info *riP)
 		riP->r_length = 1;
 		break;
 	    case 4:
+#ifdef PPC
+		if(fixP->fx_r_type == PPC_RELOC_BR14_predicted)
+		    riP->r_length = 3;
+		else
+#endif
 		riP->r_length = 2;
 		break;
+#if defined(ARCH64)
+	    case 8:
+		riP->r_length = 3;
+		break;
+#endif /* defined(ARCH64) */
 	    default:
+		layout_file = fixP->file;
+		layout_line = fixP->line;
 		as_fatal("Bad fx_size (0x%x) in fix_to_relocation_info()\n",
 			 fixP->fx_size);
 	}
@@ -969,11 +1133,15 @@ struct relocation_info *riP)
 	 * For undefined symbols this will be an external relocation entry.
 	 * Or if this is an external coalesced symbol.
 	 */
+#if defined(I386) && defined(ARCH64)
+	if (fixP->fx_subsy == NULL && !is_local_symbol(symbolP)) {
+#else
 	if((symbolP->sy_type & N_TYPE) == N_UNDF ||
 	   ((symbolP->sy_type & N_EXT) == N_EXT &&
 	    (symbolP->sy_type & N_TYPE) == N_SECT &&
 	    is_section_coalesced(symbolP->sy_other) &&
 	    fixP->fx_subsy == NULL)){
+#endif
 	    riP->r_extern = 1;
 	    riP->r_symbolnum = symbolP->sy_number;
 	}
@@ -995,6 +1163,36 @@ struct relocation_info *riP)
 	     * and the second has the subtract symbol value.
 	     */
 	    if(fixP->fx_subsy != NULL){
+#if defined(I386) && defined(ARCH64)
+		/* Encode fixP->fx_subsy (B) first, then symbolP (fixP->fx_addsy) (A). */
+		if (is_local_symbol(fixP->fx_subsy))
+		{
+			riP->r_extern = 0;
+			riP->r_symbolnum = fixP->fx_subsy->sy_other;
+		}
+		else
+		{
+			riP->r_extern = 1;
+			riP->r_symbolnum = fixP->fx_subsy->sy_number;
+		}
+		riP->r_type = X86_64_RELOC_SUBTRACTOR;
+		
+		/* Now write out the unsigned relocation entry. */
+		riP++;
+		*riP = *(riP - 1);
+		if (is_local_symbol(fixP->fx_addsy))
+		{
+			riP->r_extern = 0;
+			riP->r_symbolnum = fixP->fx_addsy->sy_other;
+		}
+		else
+		{
+			riP->r_extern = 1;
+			riP->r_symbolnum = fixP->fx_addsy->sy_number;
+		}
+		riP->r_type = X86_64_RELOC_UNSIGNED;
+		return(2 * sizeof(struct relocation_info));
+#endif
 #ifdef PPC
 		if(fixP->fx_r_type == PPC_RELOC_HI16)
 		    sectdiff = PPC_RELOC_HI16_SECTDIFF;
@@ -1002,6 +1200,8 @@ struct relocation_info *riP)
 		    sectdiff = PPC_RELOC_LO16_SECTDIFF;
 		else if(fixP->fx_r_type == PPC_RELOC_HA16)
 		    sectdiff = PPC_RELOC_HA16_SECTDIFF;
+		else if(fixP->fx_r_type == PPC_RELOC_LO14)
+		    sectdiff = PPC_RELOC_LO14_SECTDIFF;
 		else
 #endif
 #ifdef HPPA
@@ -1019,10 +1219,17 @@ struct relocation_info *riP)
 		else
 #endif
 		{
-		    if(fixP->fx_r_type != 0)
-			as_fatal("incorrect fx_r_type (%u) for fx_subsy != 0 "
-				 "in fix_to_relocation_info()",fixP->fx_r_type);
-		    sectdiff = RELOC_SECTDIFF;
+		    if(fixP->fx_r_type != 0){
+			layout_file = fixP->file;
+			layout_line = fixP->line;
+			as_fatal("Internal error: incorrect fx_r_type (%u) for "
+				 "fx_subsy != 0 in fix_to_relocation_info()",
+				 fixP->fx_r_type);
+		    }
+		    if((!(fixP->fx_addsy->sy_type & N_EXT)) && flagseen['k'])
+			sectdiff = RELOC_LOCAL_SECTDIFF;
+		    else
+			sectdiff = RELOC_SECTDIFF;
 		}
 		memset(&sri, '\0',sizeof(struct scattered_relocation_info));
 		sri.r_scattered = 1;
@@ -1036,7 +1243,8 @@ struct relocation_info *riP)
 
 		sri.r_type      = RELOC_PAIR;
 		sri.r_value     = fixP->fx_subsy->sy_value;
-		if(sectdiff == RELOC_SECTDIFF)
+		if(sectdiff == RELOC_SECTDIFF ||
+		   sectdiff == RELOC_LOCAL_SECTDIFF)
 		    sri.r_address = 0;
 #ifdef PPC
 		else if(sectdiff == PPC_RELOC_HI16_SECTDIFF ||
@@ -1045,7 +1253,8 @@ struct relocation_info *riP)
 				     fixP->fx_subsy->sy_value
 				     + fixP->fx_offset) & 0xffff;
 		}
-		else if(sectdiff == PPC_RELOC_LO16_SECTDIFF){
+		else if(sectdiff == PPC_RELOC_LO16_SECTDIFF ||
+			sectdiff == PPC_RELOC_LO14_SECTDIFF){
 		    sri.r_address = ((symbolP->sy_value -
 				      fixP->fx_subsy->sy_value +
 				      fixP->fx_offset) >> 16) & 0xffff;
@@ -1079,16 +1288,21 @@ struct relocation_info *riP)
 		return(2 * sizeof(struct relocation_info));
 	    }
 	    /*
-	     * Determine if this is left as a local relocation entry must be
+	     * Determine if this is left as a local relocation entry or must be
 	     * changed to a scattered relocation entry.  These entries allow
 	     * the link editor to scatter the contents of a section and a local
 	     * relocation can't be used when an offset is added to the symbol's
 	     * value (symbol + offset).  This is because the relocation must be
 	     * based on the value of the symbol not the value of the expression.
 	     * Thus a scattered relocation entry that encodes the address of the
-	     * symbol is used when the offset is non-zero.
+	     * symbol is used when the offset is non-zero.  Unfortunately this
+	     * encoding only allows for 24 bits in the r_address field and can
+	     * overflow.  So it if it would overflow we don't create a
+	     * scattered relocation entry and hope the offset does not reach
+	     * out of the block or the linker will not be doing scattered
+	     * loading on this symbol in this object file.
 	     */
-#if !defined(I860)
+#if !defined(I860) && !(defined(I386) && defined(ARCH64))
 	    /*
 	     * For processors that don't have all references as unique 32 bits
 	     * wide references scattered relocation entries are not generated.
@@ -1097,6 +1311,7 @@ struct relocation_info *riP)
 	     * by two references to two different symbols.
 	     */
 	    if(fixP->fx_offset != 0 &&
+	       (riP->r_address & 0xff000000) == 0 &&
 	       ((symbolP->sy_type & N_TYPE) & ~N_EXT) != N_ABS
 #ifdef M68K
 	       /*
@@ -1110,7 +1325,6 @@ struct relocation_info *riP)
 	             (fixP->fx_size == 4 && fixP->fx_offset == 4)) )
 #endif /* M68K */
 	       ){
-
 		memset(&sri, '\0',sizeof(struct scattered_relocation_info));
 		sri.r_scattered = 1;
 		sri.r_length    = riP->r_length;
@@ -1120,7 +1334,7 @@ struct relocation_info *riP)
 		sri.r_value     = symbolP->sy_value;
 		*riP = *((struct relocation_info *)&sri);
 	    }
-#endif /* !defined(I860) */
+#endif /* !defined(I860) && !(defined(I386) && defined(ARCH64)) */
 	}
 	count = 1;
 	riP++;
@@ -1169,6 +1383,11 @@ struct relocation_info *riP)
 		case 4:
 		    riP->r_length = 2;
 		    break;
+#if defined(ARCH64)
+                case 8:
+                    riP->r_length = 3;
+                    break;
+#endif /* defined(ARCH64) */
 		default:
 		    as_fatal("Bad fx_size (0x%x) in fix_to_relocation_info()\n",
 			     fixP->fx_size);
@@ -1316,3 +1535,28 @@ I860_tweeks(void)
 	clear_section_flags();
 }
 #endif
+
+/* FROM write.c line 2764 */
+void
+number_to_chars_bigendian (char *buf, signed_expr_t val, int n)
+{
+  if (n <= 0)
+    abort ();
+  while (n--)
+    {
+      buf[n] = val & 0xff;
+      val >>= 8;
+    }
+}
+
+void
+number_to_chars_littleendian (char *buf, signed_expr_t val, int n)
+{
+  if (n <= 0)
+    abort ();
+  while (n--)
+    {
+      *buf++ = val & 0xff;
+      val >>= 8;
+    }
+}
